# itunesdb.tcl
#
# Tcl library for reading iTunesDB files.
#
# Copyright (C) 2004 Andy Goth <unununium@openverse.com>
# For more information visit http://ioioio.net/devel/itunesdb/
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# See the file README for information on how to get the full text of the GNU
# General Public License.

package require snit ;# XXX: I'm not sure what minimum version I require.
package provide itunesdb 0.1.0

snit::type ::itunesdb {
    variable path       ;# Name of iTunesDB file, or "".
    variable chan       ;# Tcl channel for reading the file.

    variable state      ;# What is currently being parsed?
    variable index      ;# Next track or playlist to read.
    variable pos        ;# Current file position in bytes.

    variable track_pos  ;# File position at which first track resides.
    variable track_num  ;# Total number of tracks in the file.

    variable plist_pos  ;# File position at which first playlist resides.
    variable plist_num  ;# Total number of playlists in the file.

    # itunesdb $filename
    #
    # Opens an iTunesDB database file.
    #
    # $filename - Path to the iTunesDB file.
    #
    # return - Command through which further access is conducted ($db).
    constructor {filename} {
        set path $filename
        set chan [open $path r]

        # Treat the iTunesDB file as a raw binary--- no translations.
        fconfigure $chan -encoding binary -translation binary -eofchar ""

        set pos   0         ;# Start at the beginning of the file.
        set state initial   ;# Initial state: expect 'mhbd'.

        # Get file summary information.
        $self read
    }

    # $db destroy
    #
    # Closes the database file.
    destructor {
        if {[info exists chan]} {
            close $chan
        }
    }

    # $db count $which
    #
    # Returns the number of elements of the given type in the database file.
    #
    # $which - Element type to count.  Can be track or playlist.
    #
    # return - Number of tracks or playlists in the file.
    method count {which} {
        switch -- $which {
        track {
            return $track_num
        } playlist {
            return $plist_num
        } default {
            error "bad type \"$which\": must be track or playlist"
        }}
    }

    # $db seek $which
    #
    # Changes the current position in the database file.
    #
    # $which - Next section to read from.  Can be track or playlist.
    #
    # return - A list of the format {$section $index}, as in [$db tell].
    method seek {which} {
        switch -- $which {
        track {
            set num   $track_num
            set pos   $track_pos
            set state track_data
        } playlist {
            set num   $plist_num
            set pos   $plist_pos
            set state plist_data
        }}

        set index 0
        if {$num == 0} {
            set    state end
            return [list end    0]
        } else {
            return [list $which 0]
        }
    }

    # $db tell
    #
    # Returns the current position in the database file.
    #
    # return - A list of the format {$section $index}.
    #     $section - Current section type: other, end, track, or playlist.
    #     $index   - Index of the next record in this section.
    method tell {} {
        switch -- $state {
        initial    {return [list other    0     ]}
        end        {return [list end      0     ]}
        track_data {return [list track    $index]}
        plist_data {return [list playlist $index]}
        default    {error "BUG: invalid state \"$state\""}
        }
    }

    # $db remaining
    #
    # Returns the number of records remaining.
    #
    # return - Number of records.
    method remaining {} {
        switch -- $state {
        initial    {return 0}
        end        {return 0}
        track_data {return [expr {$track_num - $index}]}
        plist_data {return [expr {$plist_num - $index}]}
        default    {error "BUG: invalid state \"$state\""}
        }
    }

    # $db read
    #
    # Reads and returns the next record in the database file.
    #
    # return - The data read from the file.  The format depends on the most
    #     recent call to [$db seek].
    #     (initial) - {summary $filename track_count $num playlist_count $num}
    #     track     - {track $id key $val key $val ...}
    #     playlist  - {playlist $name type $type tracks {$id $id $id ...}}
    # error - "file is not an iTunes DB" - Incorrect format on first read.
    #     "no more records" - Attempt to read past end of section.
    method read {} {
        set next_pos $pos

        switch -- $state {
        initial    {set goal summary }
        track_data {set goal track   }
        plist_data {set goal playlist} 
        end        {error "no more records"              }
        default    {error "BUG: invalid state \"$state\""}
        }

        while {1} {
            # Read the header for this block.
            if {[catch {set header [$self GetVal hdr 0]} err]} {
                set state end
                if {[eof $chan]} {
                    if {$state eq "initial"} {
                        error "file is not an iTunes DB"
                    } else {
                        break
                    }
                } else {
                    error $err
                }
            }

            switch -- [catch {

            # Process this block.
            switch -glob -- "$state $header" {
            "initial mhbd" {
                # Found initial header.  Skip ahead to mhsd.
                incr next_pos  [$self GetVal 4]
                set  state     ready
            } "initial *" {
                # Bad bits at the beginning of the file.
                set  err_pos   [$self GetVal 4]
                set  err_state initial
                error "file is not an iTunes DB"
            } "ready mhsd" {
                # Top-level aggregate block.
                incr next_pos  [$self GetVal  4]
                set  next_mhsd [expr {$pos + [$self GetVal 8]}]
                set  err_state ready
                set  err_pos   $next_mhsd
                set  type      [$self GetVal 12]
                switch -- $type {
                1 {
                    # Track list.
                    if {[info exists track_pos]} {
                        error "extra track list section"
                    }
                    set state track_header
                } 2 {
                    # Playlists.
                    if {[info exists plist_pos]} {
                        error "extra playlist section"
                    }
                    set state plist_header
                } default {
                    # Neither track list nor playlist.
                    set next_pos $next_mhsd
                }}
            } "track_header mhlt" {
                # Beginning of track list.
                set track_pos [expr {$pos + [$self GetVal 4]}]
                set track_num [$self GetVal 8]
                set next_pos  $next_mhsd
                set state     ready
            } "plist_header mhlp" {
                # Beginning of playlists.
                set plist_pos [expr {$pos + [$self GetVal 4]}]
                set plist_num [$self GetVal 8]
                set next_pos  $next_mhsd
                set state     ready
            }
                
            "track_data mhit" {
                # Track entry, numeric portion.
                incr next_pos [$self GetVal 4]
                set  err_pos  [expr {$pos + [$self GetVal 8]}]

                if {$index == $track_num - 1} {
                    set err_state end
                } else {
                    set err_state track_data
                }

                set fld_num [$self GetVal 12]
                set result  [list track [$self GetVal 16]]

                foreach {field offset} {
                    rating     28     mod_time  32     file_size   36
                    duration   40     track_num 44     track_count 48
                    year       52     bit_rate  56     volume      64
                    play_count 80     last_play 88     cd_num      92
                    cd_count   96
                } {
                    lappend result $field [$self GetVal $offset]
                }

                if {$fld_num == 0} {
                    # No text fields to read, so quit now.
                    break
                } else {
                    # Expect text fields (mhod).
                    set fld_idx 0
                    set state   track_data_text
                }
            } "track_data_text mhod" {
                # Track entry, textual portion.
                incr next_pos [$self GetVal 8]

                # Determine field type.
                switch -- [$self GetVal 12] {
                 1 {set field title    }       2 {set field ipod_path}
                 3 {set field album    }       4 {set field artist   }
                 5 {set field genre    }       6 {set field type     }
                 7 {set field equalizer}       8 {set field comment  }
                12 {set field composer } default {set field unknown  }
                }

                if {$field ne "unknown"} {
                    # Read textual data.
                    lappend result $field
                    lappend result [$self GetVal utf16 40 [$self GetVal 28]]
                }

                incr fld_idx
                if {$fld_idx == $fld_num} {
                    # No more text fields to read.  Quit and return the record.
                    break
                }
            }
                
            "plist_data mhyp" {
                # Playlist, numeric portion.
                incr next_pos [$self GetVal 4]
                set  err_pos  [expr {$pos + [$self GetVal 8]}]

                if {$index == $plist_num - 1} {
                    set err_state end
                } else {
                    set err_state plist_data
                }

                # Total number of tracks in playlist.
                set trk_num [$self GetVal 16]

                # Calculate default playlist name.
                if {([$self GetVal 20] & 255) == 1} {
                    set pl_name Master-PL
                    set pl_type master
                } else {
                    set pl_name Playlist
                    set pl_type custom
                }
                set pl_name [format "%s (%08x)" $pl_name $pos]

                # Expect text fields (mhod) or playlist entries (mhyp).
                set trk_idx 0
                set result  {}
                set state   plist_data_entry
            } "plist_data_entry mhod" {
                # Playlist, textual portion.
                incr next_pos [$self GetVal 8]

                # I assume the title mhod appears somewhere before the final
                # mhip entry--- if not, oh well, the defaults are fine. :^)

                if {[$self GetVal 12] == 1} {
                    # Playlist title.
                    set pl_name [$self GetVal utf16 40 [$self GetVal 28]]
                } else {
                    # Unknown mhod type (most likely 100).  Skip.
                }
            } "plist_data_entry mhip" {
                # Playlist, track entry.
                incr  next_pos [$self GetVal  8]
                lappend result [$self GetVal 24]

                incr trk_idx
                if {$trk_idx == $trk_num} {
                    # This was the last track in the playlist.
                    break
                }
            }

            default {
                error "unexpected section \"$header\""
            }}

            } err] {
            0 {
                # TCL_OK: Success.  Go to the next block.
                set pos   $next_pos
            } 1 {
                # TCL_ERROR: Failure.  Attempt to resynchronize.
                set pos   $err_pos
                set state $err_state
                error     $err                
            } 3 {
                # TCL_BREAK: Success.  Return the completed record.
                set pos   $next_pos
                break
            }}
        }

        switch -- $goal {
        summary {
            set state end
            foreach which {Track Plist} {
                if {![info exists ${which}_pos]} {
                    set ${which}_num 0
                }
            }
            return [list summary $path track_count $track_num\
                         playlist_count $plist_num]
        } track {
            incr index
            if {$index == $track_num} {
                set state end
            } else {
                set state track_data
            }
            return $result
        } playlist {
            incr index
            if {$index == $plist_num} {
                set state end
            } else {
                set state plist_data
            }
            return [list playlist $pl_name type $pl_type tracks $result]
        }}
    }

    # $db GetVal ?type? offset ?len?
    #
    # Reads numeric or textual data from the database file.
    #
    # (Internal)
    #
    # $type - Type of data to read from file.  Can be one of the following:
    #     byte  -  8-bit signed value         (numeric)
    #     short - 16-bit signed value         (numeric)
    #     int   - 32-bit signed value         (numeric) (default)
    #     long  - 64-bit signed value         (numeric)
    #     hdr   - four-byte header identifier (numeric)
    #     bin   - binary data                 (textual)
    #     utf16 - UTF-16 data                 (textual)
    # $offset - Offset relative to $Pos at which to read data.
    # $len - Number of bytes to read.  Only valid with textual $type.
    #
    # return - The data read from the file.
    # error - "end of file" - Attempt to read past the end of the file.
    method GetVal {args} {
        switch -- [llength $args] {
        1 {
            set type   int
            set offset [lindex $args 0]
            set length ""
        } 2 {
            set type   [lindex $args 0]
            set offset [lindex $args 1]
            set length ""
        } 3 {
            set type   [lindex $args 0]
            set offset [lindex $args 1]
            set length [lindex $args 2]
        } default {
            error "wrong # args: should be \"\$db GetVal ?type? offset ?len?\""
        }}

        switch -- $type {
        byte  {set width 1; set fmt c ; set mode num}
        short {set width 2; set fmt s ; set mode num}
        int   {set width 4; set fmt i ; set mode num}
        long  {set width 8; set fmt w ; set mode num}
        hdr   {set width 4; set fmt a4; set mode num}
        bin   {set fmt binary         ; set mode txt}
        utf16 {set fmt unicode        ; set mode txt}
        default {error "invalid type \"$type\""}
        }

        switch -- $mode {
        num {
            if {$length ne ""} {
                error "cannot specify len with numeric data type"
            } else {
                seek $chan [expr {$pos + $offset}]
                if {[binary scan [read $chan $width] $fmt data] == 0} {
                    error "end of file"
                }
            }
        } txt {
            if {$length eq ""} {
                error "must specify len with textual data type"
            } else {
                seek $chan [expr {$pos + $offset}]
                set data [read $chan $length]
                if {[string length $data] != $length} {
                    error "end of file"
                }
                if {$fmt ne "binary"} {
                    set data [encoding convertfrom $fmt $data]
                }
            }
        }}

        # puts "* GetVal $type ${pos}+$offset: $data"
        return $data
    }
}

# vim: set ts=4 sts=4 sw=4 tw=80 et:
